from autogalaxy.aggregator.plane import PlaneAgg
from autogalaxy.aggregator.imaging import ImagingAgg
from autogalaxy.aggregator.fit_imaging import FitImagingAgg
from autogalaxy.aggregator.interferometer import InterferometerAgg
from autogalaxy.aggregator.fit_interferometer import FitInterferometerAgg
